from torchok.models.heads.classification.arcface_head import (
    ArcFaceHead,
)
from torchok.models.heads.classification.classification_head import (
    ClassificationHead,
)


__all__ = [
    'ArcFaceHead',
    'ClassificationHead',
]



from abc import ABC, abstractmethod
from typing import Union, Optional

import cv2
import numpy as np
from torch.utils.data import Dataset
from albumentations import BasicTransform
from albumentations.core.composition import BaseCompose


class ImageDataset(Dataset, ABC):
    """An abstract class for image dataset."""

    def __init__(self,
                 transform: Optional[Union[BasicTransform, BaseCompose]],
                 augment: Optional[Union[BasicTransform, BaseCompose]] = None,
                 image_dtype: str = 'float32',
                 grayscale: bool = False,
                 test_mode: bool = False):
        """Init ImageDataset.

        Args:
            transform: Transform to be applied on a sample. This should have the
                interface of transforms in `albumentations` library.
            augment: Optional augment to be applied on a sample.
                This should have the interface of transforms in `albumentations` library.
            image_dtype: Data type of the torch tensors related to the image.
            grayscale: If True, image will be read as grayscale otherwise as RGB.
            test_mode: If True, only image without labels will be returned.
        """
        self._test_mode = test_mode
        self._transform = transform
        self._augment = augment
        self._image_dtype = image_dtype
        self._grayscale = grayscale


    def _apply_transform(self, transform: Union[BasicTransform, BaseCompose], sample: dict) -> dict:
        """Is transformations based on API of albumentations library.

        Args:
            transform: Transformations from `albumentations` library.
                https://github.com/albumentations-team/albumentations/
            sample: Sample which the transformation will be applied to.

        Returns:
            Transformed sample.
        """
        if transform is None:
            return sample

        new_sample = transform(**sample)
        return new_sample

    def _read_image(self, image_path: str) -> np.ndarray:
        image = cv2.imread(str(image_path), int(not self._grayscale))

        if image is None:
            raise ValueError(f'{image_path} image does not exist')
        if self._grayscale:
            image = image[..., None]
        else:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        return image

    @abstractmethod
    def __len__(self) -> int:
        """Dataset length."""
        pass

    @abstractmethod
    def __getitem__(self, item: int) -> dict:
        """Get item sample."""
        pass

    @property
    def test_mode(self) -> bool:
        """Is test mode."""
        return self._test_mode

    @property
    def transform(self) -> Optional[Union[BasicTransform, BaseCompose]]:
        """Is transform to be applied on a sample."""
        return self._transform

    @property
    def augment(self) -> Optional[Union[BasicTransform, BaseCompose]]:
        """Is optional augment to be applied on a sample."""
        return self._augment

    @property
    def image_dtype(self) -> str:
        """Is data type of the torch tensors related to the image."""
        return self._image_dtype

    @property
    def grayscale(self) -> bool:
        """Is grayscale mode."""
        return self._grayscale



from . import model_checkpoint_with_onnx
from . import finalize_logger
from . import freeze_unfreeze

from pytorch_lightning.callbacks import (
    DeviceStatsMonitor, EarlyStopping, GradientAccumulationScheduler, LearningRateMonitor, ModelPruning,
    ModelSummary, QuantizationAwareTraining, RichModelSummary, RichProgressBar, StochasticWeightAveraging,
    Timer, TQDMProgressBar, BackboneFinetuning
)

from torchok.constructor import CALLBACKS


CALLBACKS.register_class(DeviceStatsMonitor)
CALLBACKS.register_class(EarlyStopping)
CALLBACKS.register_class(GradientAccumulationScheduler)
CALLBACKS.register_class(LearningRateMonitor)
CALLBACKS.register_class(ModelPruning)
CALLBACKS.register_class(ModelSummary)
CALLBACKS.register_class(QuantizationAwareTraining)
CALLBACKS.register_class(RichModelSummary)
CALLBACKS.register_class(RichProgressBar)
CALLBACKS.register_class(StochasticWeightAveraging)
CALLBACKS.register_class(Timer)
CALLBACKS.register_class(TQDMProgressBar)
CALLBACKS.register_class(BackboneFinetuning)



from pathlib import Path
from typing import Union, Optional

import re
import torch
import numpy as np
import pandas as pd
from albumentations import BasicTransform
from albumentations.core.composition import BaseCompose

from torchok.data.datasets.base import ImageDataset
from torchok.constructor import DATASETS


@DATASETS.register_class
class ImageClassificationDataset(ImageDataset):
    """A generic dataset for multilabel/multiclass image classification task.

    .. csv-table:: Multiclass task csv example.
        :header: image_path, label
        
        cat_1.jpg, 1
        dog_1.jpg, 0

    .. csv-table:: Multilabel task csv example.
        :header: image_path, label

        cat_dog_1.jpg, 0 1
        cat_dog_2.jpg, 0 1
        dog_1.jpg, 0
    """

    def __init__(self,
                 data_folder: str,
                 csv_path: str,
                 num_classes: int,
                 transform: Optional[Union[BasicTransform, BaseCompose]],
                 augment: Optional[Union[BasicTransform, BaseCompose]] = None,
                 image_dtype: str = 'float32',
                 target_dtype: str = 'long',
                 csv_columns_mapping: dict = None,
                 grayscale: bool = False,
                 test_mode: bool = False,
                 multilabel: bool = False,
                 lazy_init: bool = False):
        """Init ImageClassificationDataset.

        Args:
            data_folder: Directory with all the images.
            csv_path: Path to the csv file with path to images and annotations.
                Path to images must be under column ``input_column`` and
                annotations must be under ``target_column`` column.
            num_classes: Number of classes (i.e. maximum class index in the dataset).
            transform: Transform to be applied on a sample. This should have the
                interface of transforms in `albumentations`_ library.
            augment: Optional augment to be applied on a sample.
                This should have the interface of transforms in `albumentations`_ library.
            image_dtype: Data type of of the torch tensors related to the image.
            target_dtype: Data type of of the torch tensors related to the target.
            csv_columns_mapping: Matches mapping column names. Key - TorchOK column name, Value - csv column name.
                default value: {'image_path': 'image_path', 'label': 'label'}
            grayscale: If True, image will be read as grayscale otherwise as RGB.
            test_mode: If True, only image without labels will be returned.
            multilabel: If True, targets are being converted to multihot vector for multilabel task.
                        If False, dataset prepares targets for multiclass classification.
            lazy_init: If True, the target variable is converted to multihot when ``__getitem__`` is called (multilabel).
                       For multiclass will check the class index to fit the range when ``__getitem__`` is called.

        .. _albumentations: https://albumentations.ai/docs/
        """
        super().__init__(transform, augment, image_dtype, grayscale, test_mode)

        self.__data_folder = Path(data_folder)
        self.__num_classes = num_classes
        self.__target_dtype = target_dtype
        self.__multilabel = multilabel
        self.__lazy_init = lazy_init
        self.__csv_path = csv_path
        self.__csv_columns_mapping = csv_columns_mapping if csv_columns_mapping is not None\
            else {'image_path': 'image_path',
                  'label': 'label'}

        self.__input_column = self.__csv_columns_mapping['image_path']
        self.__target_column = self.__csv_columns_mapping['label']

        if self.__multilabel:
            self.__csv = pd.read_csv(self.__data_folder / self.__csv_path, dtype={self.__input_column: 'str',
                                                                                  self.__target_column: 'str'})
            if not self.__lazy_init and not self.test_mode:
                self.__csv[self.__target_column] = self.__csv[self.__target_column].apply(self.__process_multilabel)
        else:
            self.__csv = pd.read_csv(self.__data_folder / self.__csv_path, dtype={self.__input_column: 'str',
                                                                                  self.__target_column: 'int'})
            if not self.__lazy_init and not self.test_mode:
                self.__csv[self.__target_column] = self.__csv[self.__target_column].apply(self.__process_multiclass)

    def __getitem__(self, idx: int) -> dict:
        """Get item sample.

        Returns:
            sample: dict, where
            sample['image'] - Tensor, representing image after augmentations and transformations, dtype=image_dtype.
            sample['target'] - Target class or labels, dtype=target_dtype.
            sample['index'] - Index.
        """
        record = self.__csv.iloc[idx]
        image_path = self.__data_folder / record[self.__input_column]
        image = self._read_image(image_path)
        sample = {'image': image}
        sample = self._apply_transform(self.augment, sample)
        sample = self._apply_transform(self.transform, sample)
        sample['image'] = sample['image'].type(torch.__dict__[self._image_dtype])
        sample['index'] = idx

        if self._test_mode:
            return sample

        sample['target'] = record[self.__target_column]

        if self.__lazy_init:
            if self.__multilabel:
                sample['target'] = self.__process_multilabel(str(sample['target']))
                sample['target'] = sample['target'].astype(self.target_dtype)
            else:
                sample['target'] = self.__process_multiclass(sample['target'])

        sample['target'] = torch.tensor(sample['target']).type(torch.__dict__[self.__target_dtype])

        return sample

    def __len__(self) -> int:
        """Dataset length."""
        return len(self.__csv)

    def __process_multiclass(self, class_idx: int) -> int:
        """Check the class index to fit the range.

        Args:
            class_idx: Target class index for multiclass classification.

        Returns:
            Verified class index.

        Raises:
            ValueError: If class index is out of range.
        """
        if class_idx >= self.__num_classes:
            raise ValueError(f'Target column contains class index: {class_idx}, '
                             f'it\'s more than num_classes = {self.__num_classes}')
        return class_idx

    def __process_multilabel(self, labels: str) -> np.array:
        """Convert label to multihot representation.

        Args:
            label: Target labels for multilabel classification.
                The class indexes must be separated by any separator.

        Returns:
            Multihot vector.

        Raises:
            ValueError: If class label is out of range.
        """
        labels = list(map(int, re.findall(r'\d+', labels)))

        max_label = max(labels)

        if max_label < self.__num_classes:
            multihot = np.zeros((self.__num_classes,), dtype=bool)
            multihot[labels] = True
        else:
            raise ValueError(f'Target column contains label: {max_label}, '
                             f'it\'s more than num_classes = {self.__num_classes}')
        return multihot

    @property
    def csv_columns_mapping(self) -> dict:
        """Column name matching."""
        return self.__csv_columns_mapping

    @property
    def target_dtype(self) -> str:
        """Is target type."""
        return self.__target_dtype

    @property
    def multilabel(self) -> bool:
        """Is task mode."""
        return self.__multilabel

    @property
    def num_classes(self) -> int:
        """Is number of classes."""
        return self.__num_classes

    @property
    def lazy_init(self) -> bool:
        """Is lazy init mode."""
        return self.__lazy_init



from typing import Any, Dict, Tuple, List, Optional, Union
from abc import ABC, abstractmethod

import torch
from torch.utils.data import DataLoader
from pytorch_lightning import LightningModule
from omegaconf import DictConfig

from torchok.constructor.config_structure import Phase
from torchok.constructor.constructor import Constructor
from torchok.constructor.load import load_checkpoint


class BaseTask(LightningModule, ABC):
    """An abstract class that represent main methods of tasks."""

    def __init__(self, hparams: DictConfig):
        """Init BaseTask.

        Args:
            hparams: Hyperparameters that set in yaml file.
        """
        super().__init__()
        self.save_hyperparameters(hparams)
        self.__constructor = Constructor(hparams)
        self._input_tensors = []
        self._losses = self.__constructor.configure_losses() if hparams.get('joint_loss') is not None else None
        self._hparams = hparams
        self._metrics_manager = self.__constructor.configure_metrics_manager()
        
        # `inputs` key in yaml used for model checkpointing.
        inputs = hparams.task.params.get('inputs')
        if inputs is not None:
            for input_params in inputs:
                input_tensor = torch.rand(*input_params['shape']).type(torch.__dict__[input_params['dtype']])
                self._input_tensors.append(input_tensor)

    @abstractmethod
    def forward(self, *args, **kwargs) -> torch.Tensor:
        """Abstract forward method for validation an test."""
        pass

    @abstractmethod
    def forward_with_gt(batch: Dict[str, Any]) -> Dict[str, torch.Tensor]:
        """Abstract forward method for training(with ground truth labels)."""
        pass

    def train_modules(self) -> List[str]:
        """Create a list of 1st-level modules that need to be optimized. The method is used to apply an optimizer
        for the returned modules.

        By default, it would be self.children().

        Returns: List of modules that need to be optimized.
        """
        return self.children()

    def configure_optimizers(self) -> Tuple[List, List]:
        """Configure optimizers."""
        modules = self.train_modules()
        opt_sched_list = self.__constructor.configure_optimizers(modules)
        return opt_sched_list

    def train_dataloader(self) -> Optional[List[DataLoader]]:
        """Implement one or more PyTorch DataLoaders for training."""
        data_params = self._hparams['data'].get(Phase.TRAIN, None)

        if data_params is None:
            return None

        data_loader = self.__constructor.create_dataloaders(Phase.TRAIN)
        return data_loader

    def val_dataloader(self) -> Optional[List[DataLoader]]:
        """Implement one or multiple PyTorch DataLoaders for prediction."""
        data_params = self._hparams['data'].get(Phase.VALID, None)

        if data_params is None:
            return None

        self.__check_drop_last_params(data_params, Phase.VALID.value)

        data_loader = self.__constructor.create_dataloaders(Phase.VALID)
        return data_loader

    def test_dataloader(self) -> Optional[List[DataLoader]]:
        """Implement one or multiple PyTorch DataLoaders for testing."""
        data_params = self._hparams['data'].get(Phase.TEST, None)

        if data_params is None:
            return None

        self.__check_drop_last_params(data_params, Phase.TEST.value)

        data_loader = self.__constructor.create_dataloaders(Phase.TEST)
        return data_loader

    def predict_dataloader(self) -> Optional[List[DataLoader]]:
        """Implement one or multiple PyTorch DataLoaders for prediction."""
        data_params = self._hparams['data'].get(Phase.PREDICT, None)

        if data_params is None:
            return None

        self.__check_drop_last_params(data_params, Phase.PREDICT.value)

        data_loader = self.__constructor.create_dataloaders(Phase.PREDICT)
        return data_loader

    def __check_drop_last_params(self, data_params: List[Dict[str, Any]], phase: str) -> None:
        for data_param in data_params:
            drop_last = data_param['dataloader'].get('drop_last', False)
            if drop_last:
                # TODO: create logger and print a warning instead
                raise ValueError(f'DataLoader parameters `drop_last` must be False in {phase} phase.')

    def on_train_start(self) -> None:
        if self.current_epoch == 0:
            load_checkpoint(self, base_ckpt_path=self._hparams.task.base_checkpoint, 
                            overridden_name2ckpt_path=self._hparams.task.overridden_checkpoints,
                            exclude_keys=self._hparams.task.exclude_keys)

    def on_test_start(self) -> None:
        load_checkpoint(self, base_ckpt_path=self._hparams.task.base_checkpoint, 
                        overridden_name2ckpt_path=self._hparams.task.overridden_checkpoints,
                        exclude_keys=self._hparams.task.exclude_keys)

    def training_step(self, batch: Dict[str, Union[torch.Tensor, int]], batch_idx: int) -> Dict[str, torch.Tensor]:
        """Complete training loop."""
        output = self.forward_with_gt(batch[0])
        total_loss, tagged_loss_values = self._losses(**output)
        self._metrics_manager.forward(Phase.TRAIN, **output)
        output_dict = {'loss': total_loss}
        output_dict.update(tagged_loss_values)
        return output_dict

    def validation_step(self, batch: Dict[str, Union[torch.Tensor, int]], batch_idx: int) -> Dict[str, torch.Tensor]:
        """Complete validation loop."""
        output = self.forward_with_gt(batch)
        self._metrics_manager.forward(Phase.VALID, **output)

        # In arcface classification task, if we try to compute loss on test dataset with different number
        # of classes we will crash the train study.
        if self._hparams.task.compute_loss_on_valid:
            total_loss, tagged_loss_values = self._losses(**output)
            output_dict = {'loss': total_loss}
            output_dict.update(tagged_loss_values)
        else:
            output_dict = {}
        
        return output_dict

    def test_step(self, batch: Dict[str, Union[torch.Tensor, int]], batch_idx: int) -> None:
        """Complete test loop."""
        output = self.forward_with_gt(batch)
        self._metrics_manager.forward(Phase.TEST, **output)

    def training_step_end(self, outputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        output_dict = {tag: value.mean() for tag, value in self.all_gather(outputs, sync_grads=True).items()}

        for tag, value in output_dict.items():
            self.log(f'train/{tag}', value, on_step=True, on_epoch=False)

        return output_dict

    def validation_step_end(self, outputs: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        output_dict = {tag: value.mean() for tag, value in self.all_gather(outputs).items()}

        for tag, value in output_dict.items():
            self.log(f'valid/{tag}', value, on_step=True, on_epoch=False)

        return output_dict

    def training_epoch_end(self,
                           training_step_outputs: List[Dict[str, torch.Tensor]]) -> None:
        """It's calling at the end of the training epoch with the outputs of all training steps."""
        self.log_dict(self._metrics_manager.on_epoch_end(Phase.TRAIN))

    def validation_epoch_end(self,
                             valid_step_outputs: List[Dict[str, torch.Tensor]]) -> None:
        """It's calling at the end of the validation epoch with the outputs of all validation steps."""
        self.log_dict(self._metrics_manager.on_epoch_end(Phase.VALID))

    def test_epoch_end(self,
                       test_step_outputs: List[Dict[str, torch.Tensor]]) -> None:
        """It's calling at the end of a test epoch with the output of all test steps."""
        self.log_dict(self._metrics_manager.on_epoch_end(Phase.TEST))

    @property
    def hparams(self) -> DictConfig:
        """Hyperparameters that set in yaml file."""
        return self._hparams

    @property
    def metrics_manager(self):
        """Metrics manager."""
        return self._metrics_manager

    @property
    def losses(self):
        """Losses."""
        return self._losses

    @property
    def input_shapes(self) -> List[List[int]]:
        """Input shapes."""
        return self.__input_shapes

    @property
    def input_dtypes(self) -> List[str]:
        """Input dtypes."""
        return self.__input_dtypes

    @property
    def input_tensors(self) -> List[torch.Tensor]:
        """Input tensors."""
        return self._input_tensors



from albumentations import *
from albumentations.pytorch import *
from albumentations import BasicTransform
from albumentations.core.composition import Compose

from torchok.data.transforms.pixelwise import *
from torchok.data.transforms.spatial import *

from torchok.constructor import TRANSFORMS


TRANSFORMS.register_class(Compose)
TRANSFORMS.register_class(OneOf)
TRANSFORMS.register_class(ToTensorV2)
TRANSFORMS.register_class(Normalize)
TRANSFORMS.register_class(Resize)

TRANSFORMS.register_class(HorizontalFlip)
TRANSFORMS.register_class(VerticalFlip)
TRANSFORMS.register_class(Affine)
TRANSFORMS.register_class(CenterCrop)
TRANSFORMS.register_class(CoarseDropout)
TRANSFORMS.register_class(Crop)
TRANSFORMS.register_class(CropAndPad)
TRANSFORMS.register_class(CropNonEmptyMaskIfExists)
TRANSFORMS.register_class(ElasticTransform)
TRANSFORMS.register_class(Flip)
TRANSFORMS.register_class(GridDistortion)
TRANSFORMS.register_class(GridDropout)
TRANSFORMS.register_class(Lambda)
TRANSFORMS.register_class(LongestMaxSize)
TRANSFORMS.register_class(MaskDropout)
TRANSFORMS.register_class(NoOp)
TRANSFORMS.register_class(OpticalDistortion)
TRANSFORMS.register_class(PadIfNeeded)
TRANSFORMS.register_class(Perspective)
TRANSFORMS.register_class(PiecewiseAffine)
TRANSFORMS.register_class(PixelDropout)
TRANSFORMS.register_class(RandomCrop)
TRANSFORMS.register_class(RandomCropNearBBox)
TRANSFORMS.register_class(RandomGridShuffle)
TRANSFORMS.register_class(RandomResizedCrop)
TRANSFORMS.register_class(RandomRotate90)
TRANSFORMS.register_class(RandomScale)
TRANSFORMS.register_class(RandomSizedBBoxSafeCrop)
TRANSFORMS.register_class(RandomSizedCrop)
TRANSFORMS.register_class(Rotate)
TRANSFORMS.register_class(SafeRotate)
TRANSFORMS.register_class(ShiftScaleRotate)
TRANSFORMS.register_class(SmallestMaxSize)
TRANSFORMS.register_class(Transpose)

TRANSFORMS.register_class(AdvancedBlur)
TRANSFORMS.register_class(Blur)
TRANSFORMS.register_class(CLAHE)
TRANSFORMS.register_class(ChannelDropout)
TRANSFORMS.register_class(ChannelShuffle)
TRANSFORMS.register_class(ColorJitter)
TRANSFORMS.register_class(Downscale)
TRANSFORMS.register_class(Emboss)
TRANSFORMS.register_class(Equalize)
TRANSFORMS.register_class(FDA)
TRANSFORMS.register_class(FancyPCA)
TRANSFORMS.register_class(FromFloat)
TRANSFORMS.register_class(GaussNoise)
TRANSFORMS.register_class(GaussianBlur)
TRANSFORMS.register_class(GlassBlur)
TRANSFORMS.register_class(HistogramMatching)
TRANSFORMS.register_class(HueSaturationValue)
TRANSFORMS.register_class(ISONoise)
TRANSFORMS.register_class(ImageCompression)
TRANSFORMS.register_class(InvertImg)
TRANSFORMS.register_class(MedianBlur)
TRANSFORMS.register_class(MotionBlur)
TRANSFORMS.register_class(MultiplicativeNoise)
TRANSFORMS.register_class(PixelDistributionAdaptation)
TRANSFORMS.register_class(Posterize)
TRANSFORMS.register_class(RGBShift)
TRANSFORMS.register_class(RandomBrightnessContrast)
TRANSFORMS.register_class(RandomFog)
TRANSFORMS.register_class(RandomGamma)
TRANSFORMS.register_class(RandomRain)
TRANSFORMS.register_class(RandomShadow)
TRANSFORMS.register_class(RandomSnow)
TRANSFORMS.register_class(RandomSunFlare)
TRANSFORMS.register_class(RandomToneCurve)
TRANSFORMS.register_class(RingingOvershoot)
TRANSFORMS.register_class(Sharpen)
TRANSFORMS.register_class(Solarize)
TRANSFORMS.register_class(Superpixels)
TRANSFORMS.register_class(TemplateTransform)
TRANSFORMS.register_class(ToFloat)
TRANSFORMS.register_class(ToGray)
TRANSFORMS.register_class(ToSepia)
TRANSFORMS.register_class(UnsharpMask)



from typing import List, Tuple, Union

import torch.nn as nn
from timm.models.resnet import Bottleneck
from torch import Tensor

from torchok.constructor import NECKS
from torchok.models.modules.bricks.convbnact import ConvBnAct
from torchok.models.necks.base_neck import BaseNeck


@NECKS.register_class
class HRNetClassificationNeck(BaseNeck):
    """HRNet neck for classification task."""

    def __init__(self, in_channels: Union[List[int], Tuple[int, ...]]):
        """Init HRNetClassificationNeck.

        Args:
            in_channels: Input channels.
        """
        out_channels = 2048
        super().__init__(in_channels, out_channels)
        self.incre_modules, self.downsamp_modules, self.final_layer = self.__make_neck(in_channels)

    def __make_neck(self, in_channels: Union[List[int], Tuple[int, ...]]):
        head_block = Bottleneck
        self.head_channels = [32, 64, 128, 256]

        # Increasing the #channels on each resolution
        # from C, 2C, 4C, 8C to 128, 256, 512, 1024
        incre_modules = []
        for i, channels in enumerate(in_channels):
            incre_modules.append(self.__make_layer(head_block, channels, self.head_channels[i], 1, stride=1))
        incre_modules = nn.ModuleList(incre_modules)

        # downsampling modules
        downsamp_modules = []
        for i in range(len(in_channels) - 1):
            in_channels = self.head_channels[i] * head_block.expansion
            out_channels = self.head_channels[i + 1] * head_block.expansion
            downsamp_module = ConvBnAct(in_channels=in_channels,
                                        out_channels=out_channels,
                                        kernel_size=3,
                                        padding=1,
                                        stride=2)
            downsamp_modules.append(downsamp_module)
        downsamp_modules = nn.ModuleList(downsamp_modules)

        final_layer = ConvBnAct(in_channels=self.head_channels[3] * head_block.expansion,
                                out_channels=self.out_channels,
                                kernel_size=1,
                                padding=0,
                                stride=1)

        return incre_modules, downsamp_modules, final_layer

    def __make_layer(self, block, inplanes, planes, blocks, stride=1):
        downsample = None

        if stride != 1 or inplanes != planes * block.expansion:
            downsample = ConvBnAct(in_channels=inplanes,
                                   out_channels=planes * block.expansion,
                                   kernel_size=1,
                                   padding=0,
                                   stride=stride,
                                   bias=False,
                                   act_layer=None)

        layers = [block(inplanes, planes, stride, downsample)]
        inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(inplanes, planes))

        return nn.Sequential(*layers)

    def forward(self, x: List[Tensor]) -> Tensor:
        """Forward method."""
        y = self.incre_modules[0](x[0])
        for i in range(len(self.downsamp_modules)):
            y = self.downsamp_modules[i](y)
            if i + 1 < len(x):
                y = self.incre_modules[i + 1](x[i + 1])
        y = self.final_layer(y)
        return y



from torchok.constructor.registry import Registry


DATASETS = Registry('datasets')
TRANSFORMS = Registry('transforms')
OPTIMIZERS = Registry('optimizers')
SCHEDULERS = Registry('schedulers')
LOSSES = Registry('losses')
METRICS = Registry('metrics')
CALLBACKS = Registry('callbacks')
TASKS = Registry('tasks')
BACKBONES = Registry('backbones')
POOLINGS = Registry('poolings')
HEADS = Registry('heads')
NECKS = Registry('necks')


__all__ = [
    'DATASETS',
    'TRANSFORMS',
    'OPTIMIZERS',
    'SCHEDULERS',
    'LOSSES',
    'METRICS',
    'CALLBACKS',
    'TASKS',
    'BACKBONES',
    'POOLINGS',
    'HEADS',
    'NECKS',
]



"""TorchOK Swin Transformer V2
A PyTorch implementation of `Swin Transformer V2: Scaling Up Capacity and Resolution`
    - https://arxiv.org/abs/2111.09883

Adapted from https://github.com/microsoft/Swin-Transformer/blob/main/models/swin_transformer_v2.py
(Copyright (c) 2022 Microsoft)
and from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/swin_transformer_v2.py
(Copyright 2022, Ross Wightman)
Licensed under Apache License 2.0 [see LICENSE for details]
"""

from typing import List

import torch
import torch.nn as nn
import torch.utils.checkpoint as checkpoint
from timm.data import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from timm.models.helpers import build_model_with_cfg
from timm.models.layers import trunc_normal_
from timm.models.swin_transformer_v2 import BasicLayer as SwinBasicLayer, checkpoint_filter_fn, PatchEmbed, PatchMerging

from torchok.constructor import BACKBONES
from torchok.models.backbones import BaseBackbone


def _cfg(url='', **kwargs):
    return {
        'url': url, 'input_size': (3, 224, 224), 'pool_size': None,
        'crop_pct': .9, 'interpolation': 'bicubic', 'fixed_input_size': True,
        'mean': IMAGENET_DEFAULT_MEAN, 'std': IMAGENET_DEFAULT_STD,
        'first_conv': 'patch_embed.proj', 'classifier': 'head',
        **kwargs
    }


url_base_path = 'https://github.com/SwinTransformer/storage/releases/download/v2.0.0/'
default_cfgs = {
    'swinv2_tiny_window8_256': _cfg(url=f'{url_base_path}/swinv2_tiny_patch4_window8_256.pth',
                                    input_size=(3, 256, 256)),
    'swinv2_tiny_window16_256': _cfg(url=f'{url_base_path}/swinv2_tiny_patch4_window16_256.pth',
                                     input_size=(3, 256, 256)),
    'swinv2_small_window8_256': _cfg(url=f'{url_base_path}/swinv2_small_patch4_window8_256.pth',
                                     input_size=(3, 256, 256)),
    'swinv2_small_window16_256': _cfg(url=f'{url_base_path}/swinv2_small_patch4_window16_256.pth',
                                      input_size=(3, 256, 256)),
    'swinv2_base_window8_256': _cfg(url=f'{url_base_path}/swinv2_base_patch4_window8_256.pth',
                                    input_size=(3, 256, 256)),
    'swinv2_base_window16_256': _cfg(url=f'{url_base_path}/swinv2_base_patch4_window16_256.pth',
                                     input_size=(3, 256, 256)),

    'swinv2_base_window12_192_22k': _cfg(url=f'{url_base_path}/swinv2_base_patch4_window12_192_22k.pth',
                                         num_classes=21841, input_size=(3, 192, 192)),
    'swinv2_base_window12to16_192to256_22kft1k': _cfg(
        url=f'{url_base_path}/swinv2_base_patch4_window12to16_192to256_22kto1k_ft.pth',
        input_size=(3, 256, 256)),
    'swinv2_base_window12to24_192to384_22kft1k': _cfg(
        url=f'{url_base_path}/swinv2_base_patch4_window12to24_192to384_22kto1k_ft.pth',
        input_size=(3, 384, 384), crop_pct=1.0),
    'swinv2_large_window12_192_22k': _cfg(url=f'{url_base_path}/swinv2_large_patch4_window12_192_22k.pth',
                                          num_classes=21841, input_size=(3, 192, 192)),
    'swinv2_large_window12to16_192to256_22kft1k': _cfg(
        url=f'{url_base_path}/swinv2_large_patch4_window12to16_192to256_22kto1k_ft.pth',
        input_size=(3, 256, 256)),
    'swinv2_large_window12to24_192to384_22kft1k': _cfg(
        url=f'{url_base_path}/swinv2_large_patch4_window12to24_192to384_22kto1k_ft.pth',
        input_size=(3, 384, 384), crop_pct=1.0),
}


class BasicLayer(SwinBasicLayer):
    """ A basic Swin Transformer layer for one stage adapted for intermediate feature extraction.
    """

    def forward(self, x):
        for blk in self.blocks:
            if self.grad_checkpointing and not torch.jit.is_scripting():
                x = checkpoint.checkpoint(blk, x)
            else:
                x = blk(x)
        return self.downsample(x), x


class SwinTransformerV2(BaseBackbone):
    r""" Swin Transformer V2
        A PyTorch impl of : `Swin Transformer V2: Scaling Up Capacity and Resolution`
            - https://arxiv.org/abs/2111.09883
        Args:
            img_size: Input image size.
            patch_size: Patch size.
            in_channels: Number of input image channels.
            embed_dim: Patch embedding dimension.
            depths: Depth of each Swin Transformer layer.
            num_heads: Number of attention heads in different layers.
            window_size: Window size.
            mlp_ratio: Ratio of mlp hidden dim to embedding dim.
            qkv_bias: If True, add a learnable bias to query, key, value.
            drop_rate: Dropout rate.
            attn_drop_rate: Attention dropout rate.
            drop_path_rate: Stochastic depth rate.
            norm_layer: Normalization layer.
            ape: If True, add absolute position embedding to the patch embedding.
            patch_norm: If True, add normalization after patch embedding.
            pretrained_window_sizes: Pretrained window sizes of each layer.
    """

    def __init__(
            self, img_size: int = 224, patch_size: int = 4, in_channels: int = 3,
            embed_dim: int = 96, depths: List[int] = (2, 2, 6, 2), num_heads: List[int] = (3, 6, 12, 24),
            window_size: int = 7, mlp_ratio: float = 4., qkv_bias: bool = True,
            drop_rate: float = 0., attn_drop_rate: float = 0., drop_path_rate: float = 0.1,
            norm_layer: nn.Module = nn.LayerNorm, ape: bool = False, patch_norm: bool = True,
            pretrained_window_sizes: List[int] = (0, 0, 0, 0)):
        super().__init__(in_channels=in_channels)
        self.img_size = img_size
        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.ape = ape
        self.patch_norm = patch_norm
        self.encoder_channels = [int(embed_dim * 2 ** i) for i in range(self.num_layers)]
        self._out_channels = self.encoder_channels[-1]
        self._out_encoder_channels = self.encoder_channels

        # split image into non-overlapping patches
        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=in_channels, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)

        self.input_resolutions = []
        for i_layer in range(self.num_layers):
            resolution = (
                self.patch_embed.grid_size[0] // (2 ** i_layer),
                self.patch_embed.grid_size[1] // (2 ** i_layer)
            )
            self.input_resolutions.append(resolution)

        self.patches_resolution = self.patch_embed.grid_size

        # absolute position embedding
        if self.ape:
            self.absolute_pos_embed = nn.Parameter(torch.zeros(1, self.patch_embed.num_patches, embed_dim))
            trunc_normal_(self.absolute_pos_embed, std=.02)
        else:
            self.absolute_pos_embed = None

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule

        # build layers
        self.layers = nn.ModuleList()
        for i_layer in range(self.num_layers):
            layer = BasicLayer(
                dim=int(embed_dim * 2 ** i_layer),
                input_resolution=(
                    self.patch_embed.grid_size[0] // (2 ** i_layer),
                    self.patch_embed.grid_size[1] // (2 ** i_layer)),
                depth=depths[i_layer],
                num_heads=num_heads[i_layer],
                window_size=window_size,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                drop=drop_rate, attn_drop=attn_drop_rate,
                drop_path=dpr[sum(depths[:i_layer]):sum(depths[:i_layer + 1])],
                norm_layer=norm_layer,
                downsample=PatchMerging if (i_layer < self.num_layers - 1) else None,
                pretrained_window_size=pretrained_window_sizes[i_layer]
            )
            self.layers.append(layer)

        self.feature_norms = nn.ModuleList([norm_layer(chs) for i, chs in enumerate(self.encoder_channels)])

        self.apply(self._init_weights)
        for bly in self.layers:
            bly._init_respostnorm()

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    # Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/swin_transformer_v2.py
    # Copyright 2019 Ross Wightman
    # Licensed under The Apache 2.0 License [see LICENSE for details]
    @torch.jit.ignore
    def no_weight_decay(self) -> List[str]:
        """Create modules names for which weights decay will not be use. See BaseBackbone for more information.
        
        Returns:
            nod: Module names for which weights decay will not be use.
        """
        nod = ['absolute_pos_embed']
        for n, m in self.named_modules():
            if any([kw in n for kw in ("cpb_mlp", "logit_scale", 'relative_position_bias_table')]):
                nod.append(n)
        return nod

    def _forward_patch_emb(self, x: torch.Tensor) -> torch.Tensor:
        """Run patch embedding part.

        Args:
            x: Input tensor.

        Returns:
            x: Patch embeddings with dropout - self.pos_drop.
        """
        x = self.patch_embed(x)
        if self.ape:
            x = x + self.absolute_pos_embed
        x = self.pos_drop(x)
        return x

    def _normalize_with_bhwc_reshape(self, x: torch.Tensor, layer_number: int, normalize: bool = True) -> torch.Tensor:
        """Convert SWin BLC shape to BCHW.

        Args:
            x: Input tensor.
            layer_number: Number os SWin layer.
            normalize: If `True` do normalization.

        Returns:
            x: Tensor with shape BCHW.
        """
        # B L C
        if normalize:
            x = self.feature_norms[layer_number](x)
        H = self.input_resolutions[layer_number][0]
        W = self.input_resolutions[layer_number][1]
        C = self.encoder_channels[layer_number]
        # B C H W
        x = x.view(-1, H, W, C).permute(0, 3, 1, 2).contiguous()
        return x

    def forward_features(self, x: torch.Tensor) -> List[torch.Tensor]:
        features = [x]
        downsample_attn = self._forward_patch_emb(x)

        for i, layer in enumerate(self.layers):
            downsample_attn, attn = layer(downsample_attn)
            feature = self._normalize_with_bhwc_reshape(attn, layer_number=i, normalize=True)
            features.append(feature)

        return features

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self._forward_patch_emb(x)
        for layer in self.layers:
            x, _ = layer(x)
        x = self._normalize_with_bhwc_reshape(x, -1)
        return x


def create_swin_transformer_v2(variant, pretrained=False, **kwargs):
    kwargs_filter = ('num_classes', 'global_pool', 'in_chans')
    model = build_model_with_cfg(SwinTransformerV2, variant, pretrained, pretrained_strict=False,
                                 kwargs_filter=kwargs_filter, pretrained_filter_fn=checkpoint_filter_fn, **kwargs)
    return model


@BACKBONES.register_class
def swinv2_tiny_window16_256(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=16, embed_dim=96, depths=(2, 2, 6, 2), num_heads=(3, 6, 12, 24), **kwargs)
    return create_swin_transformer_v2('swinv2_tiny_window16_256', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_tiny_window8_256(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=8, embed_dim=96, depths=(2, 2, 6, 2), num_heads=(3, 6, 12, 24), **kwargs)
    return create_swin_transformer_v2('swinv2_tiny_window8_256', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_small_window16_256(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=16, embed_dim=96, depths=(2, 2, 18, 2), num_heads=(3, 6, 12, 24), **kwargs)
    return create_swin_transformer_v2('swinv2_small_window16_256', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_small_window8_256(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=8, embed_dim=96, depths=(2, 2, 18, 2), num_heads=(3, 6, 12, 24), **kwargs)
    return create_swin_transformer_v2('swinv2_small_window8_256', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_base_window16_256(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=16, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32), **kwargs)
    return create_swin_transformer_v2('swinv2_base_window16_256', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_base_window8_256(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=8, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32), **kwargs)
    return create_swin_transformer_v2('swinv2_base_window8_256', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_base_window12_192_22k(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=12, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32), **kwargs)
    return create_swin_transformer_v2('swinv2_base_window12_192_22k', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_base_window12to16_192to256_22kft1k(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=16, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32),
        pretrained_window_sizes=(12, 12, 12, 6), **kwargs)
    return create_swin_transformer_v2(
        'swinv2_base_window12to16_192to256_22kft1k', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_base_window12to24_192to384_22kft1k(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=24, embed_dim=128, depths=(2, 2, 18, 2), num_heads=(4, 8, 16, 32),
        pretrained_window_sizes=(12, 12, 12, 6), **kwargs)
    return create_swin_transformer_v2(
        'swinv2_base_window12to24_192to384_22kft1k', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_large_window12_192_22k(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=12, embed_dim=192, depths=(2, 2, 18, 2), num_heads=(6, 12, 24, 48), **kwargs)
    return create_swin_transformer_v2('swinv2_large_window12_192_22k', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_large_window12to16_192to256_22kft1k(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=16, embed_dim=192, depths=(2, 2, 18, 2), num_heads=(6, 12, 24, 48),
        pretrained_window_sizes=(12, 12, 12, 6), **kwargs)
    return create_swin_transformer_v2(
        'swinv2_large_window12to16_192to256_22kft1k', pretrained=pretrained, **model_kwargs)


@BACKBONES.register_class
def swinv2_large_window12to24_192to384_22kft1k(pretrained=False, **kwargs):
    """
    """
    model_kwargs = dict(
        window_size=24, embed_dim=192, depths=(2, 2, 18, 2), num_heads=(6, 12, 24, 48),
        pretrained_window_sizes=(12, 12, 12, 6), **kwargs)
    return create_swin_transformer_v2(
        'swinv2_large_window12to24_192to384_22kft1k', pretrained=pretrained, **model_kwargs)
